import unittest
import numpy as np

from pysprint.core._preprocess import *


class TestEdit(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_savgol(self):
        x, y, v, w = np.loadtxt("test_arms.txt", delimiter=",", unpack=True)
        a, b = savgol(x, y, v, w, window=10, order=3)
        c, d = savgol(x, y, v, w, window=11, order=3)
        assert len(a) == len(b)
        assert len(d) == len(d)
        with self.assertRaises(ValueError):
            savgol(x, y, v, w, window=1, order=3)

    def test_peak(self):
        x, y = np.loadtxt("test_peak.txt", delimiter=",", unpack=True)
        a, b, c, d = find_peak(x, y, [], [], threshold=0.01, pro_min=0.5, pro_max=0.5)
        assert len(a) == len(b)
        assert len(c) == len(d)
        for val in b:
            assert abs(val) > 0.01
        for val in d:
            assert abs(val) > 0.01

    def test_interpolate(self):
        x, y, v, w = np.loadtxt("test_arms.txt", delimiter=",", unpack=True)
        a, b = interpolate_data(x, y, v, w)
        assert len(a) == len(b)
        np.random.seed(1000)
        idx1 = np.random.randint(0, len(a))
        idx2 = np.random.randint(0, len(a))
        i = abs(a[idx1] - a[idx1 - 1])
        j = abs(a[idx2] - a[idx2 - 1])
        np.testing.assert_almost_equal(i, j)
        with self.assertRaises(TypeError):
            interpolate_data({"a": "1", "b": "2"}, x, y, v, window=1, order=3)

    def test_cut(self):
        x, y, v, w = np.loadtxt("test_arms.txt", delimiter=",", unpack=True)
        a, b = cut_data(x, y, v, w, start=2.2, stop=2.8)
        assert len(a) == len(b)
        np.testing.assert_almost_equal(min(a), 2.2, decimal=2)
        np.testing.assert_almost_equal(max(a), 2.8, decimal=2)
        with self.assertRaises(ValueError):
            cut_data(x, [], v, w)

    def test_convolution(self):
        x, y, v, w = np.loadtxt("test_arms.txt", delimiter=",", unpack=True)
        a, b = convolution(x, y, v, w, len(x), standev=200)
        assert len(a) == len(b)
        with self.assertRaises(ValueError):
            convolution(x, [], v, w, 10, standev=-541)


if __name__ == "__main__":
    unittest.main()



import sys
from pathlib import Path
from typing import IO, AnyStr, Union
from textwrap import dedent
import datetime
import re

import numpy as np

from pysprint.core.phase import Phase
from pysprint.config import _get_config_value
from pysprint.utils.exceptions import NotCalculatedException
from pysprint.utils import run_from_ipython

__all__ = ["defaultcallback", "eager_executor"]


PathOrBuffer = Union[str, Path, IO[AnyStr]]


class LogWriter:
    LABELS = ("GD", "GDD", "TOD", "FOD", "QOD", "SOD")

    def __init__(self, file: PathOrBuffer, phase: Phase, verbosity: Union[int, None] = None):
        self.file = file

        if not self.file.endswith((".log", ".txt")):
            self.file += ".log"


        self.phase = phase
        self.verbosity = verbosity or _get_config_value("verbosity")

    def _write_content(self, content):
        with open(self.file, "a") as logfile:
            logfile.write(content)

    def _prepare_content(self):

        if self.phase.coef_array is None:
            raise NotCalculatedException

        precision = _get_config_value("precision")

        iter_num = len(self.phase.x)

        output = dedent(f"""
        ---------------------------------------------------------------------------------------
        Date: {datetime.datetime.now()}

        Datapoints used: {iter_num}

        R^2: {self.phase._get_r_squared():.{precision}f}

        Results:
        """)

        for i, (label, value) in enumerate(zip(self.LABELS, self.phase.coef_array)):
            if value is not None and value != 0:
                output += f"{label} = {value:.{precision}f} fs^{i + 1}\n"

        if self.verbosity > 0:
            with np.printoptions(
                    threshold=sys.maxsize,
                    linewidth=np.inf,
                    precision=precision,
            ):
                output +=  f'''
                Values:
                x: {np.array2string(self.phase.x, separator=", ")}
                y: {np.array2string(self.phase.y, separator=", ")}
                '''
        return re.sub('^\s+', '', output, flags=re.MULTILINE)

    def write(self):
        content = self._prepare_content()
        self._write_content(content)


def defaultcallback(broadcaster, listener=None):
    """
    The default recorder for SPP data.
    """
    if listener is not None:
        listener._container[broadcaster] = broadcaster.emit()


def eager_executor(reference_point=None, order=None, logfile=None, verbosity=None, cleanup=True, **kwds):
    """
    This function can be used as a `callback` function. It will
    constantly monitor for SPP data changes and will re-calculate
    the dispersion on any modification.

    Parameters
    ----------
    reference_point : float
        The reference point to use.
    order : int
        The dispersion order to look for.
    cleanup : bool, optional
        Whether to flush the output and override previous results.
    logfile : string or None, optional
        If given, a logfile will be created each iteration.
    verbosity : int or None, optional
        The verbosity level for the logfile. If 0 or None then it
        won't write the data used for fitting, else it will.
    """
    if not reference_point or not order:
        raise RuntimeError(
            "Reference point and order is required."
        )

    def inner(broadcaster, listener=None):
        defaultcallback(broadcaster, listener)
        if run_from_ipython():
            from IPython.display import clear_output  # noqa
        else:
            import sys
            clear_output = sys.stdout.flush
        try:
            if cleanup:
                clear_output()
            listener.calculate(reference_point, order, **kwds)
            try:
                if logfile:
                    LogWriter(logfile, listener.GD, verbosity).write()
            except NotCalculatedException:
                pass
        except (TypeError, ValueError):
            pass

    return inner



"""
This is a utility module for RANSAC regression filtering. We use scikit-learn's RANSACRegressor class
with a custom polynomial model.

References:
https://en.wikipedia.org/wiki/Random_sample_consensus
https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.RANSACRegressor.html
"""
import warnings

import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics import mean_squared_error
from sklearn.linear_model import RANSACRegressor

warnings.simplefilter('ignore', np.RankWarning)

def make_regression_class(degree):
    class PolynomialModel:
        def __init__(self, degree=degree, coeffs=None):
            self.degree = degree
            self.coeffs = coeffs

        def fit(self, X, y):
            self.coeffs = np.polyfit(X.ravel(), y, self.degree)

        def get_params(self, deep=False):
            return {'coeffs': self.coeffs}

        def set_params(self, coeffs=None, random_state=None):
            self.coeffs = coeffs

        def predict(self, X):
            poly_eqn = np.poly1d(self.coeffs)
            y_hat = poly_eqn(X.ravel())
            return y_hat

        def score(self, X, y):
            return mean_squared_error(y, self.predict(X))
    return PolynomialModel


def run_regressor(phase, degree, plot=True, **kwds):
    # this is the most important argument for us, so we isolate it
    residual_threshold = kwds.pop("residual_threshold", 0.5) * np.std(phase.y)

    # we must create the class on the fly, because sklearn copies the class,
    # and it uses the default degree param.
    model = make_regression_class(degree=degree)()
    ransac = RANSACRegressor(
        model,
        residual_threshold=residual_threshold,
        **kwds
    )
    ransac.fit(phase.x[:, np.newaxis], phase.y)
    inlier_mask = ransac.inlier_mask_

    y_hat = ransac.predict(phase.x[:, np.newaxis])

    if plot:
        _, ax = plt.subplots()
        ax.clear()
        phase.plot(ax=ax)
        ax.plot(phase.x[inlier_mask], phase.y[inlier_mask], "k+", label="inliers")
        ax.plot(phase.x, y_hat, 'g--', label='RANSAC estimated curve')
    return phase.x[inlier_mask], phase.y[inlier_mask]



from pysprint import config as cfg


def is_nonnegative_int(value):
    if isinstance(value, int) and value >= 0:
        return True
    raise ValueError


def is_normalized_float(value):
    if isinstance(value, (float, int)):
        if 0 <= value <= 1:
            return True
        raise ValueError
    raise ValueError


_precision_doc = """
: int
    The precision of floating point numbers in the output.
"""

cfg.register_config_value("precision", 5, doc=_precision_doc, validator=is_nonnegative_int)

_verbosity_doc = """
: int
    The verbosity level of logging outputs when using eager_executor.
"""

cfg.register_config_value("verbosity", 0, doc=_verbosity_doc, validator=is_nonnegative_int)

_scan_threshold_doc = """
: float
    Determines the maximum relative distance between any calculated dispersion coefficient.
    If the relative distance is bigger than this value, the two calculated values will be
    treated as distinct: the two sides won't be averaged out.
"""

cfg.register_config_value("scan_threshold", 0.5, doc=_scan_threshold_doc, validator=is_normalized_float)



from glob import glob

import numpy as np

__all__ = ["correct_sign", "from_pat"]


# TODO: rewrite or deprecate this.
def correct_sign(x, flip_increasing=True):
    if not isinstance(x, np.ndarray):
        x = np.asarray(x)
    increasing_idx = np.where(np.diff(x) > 0)
    # inc_len = len(increasing_idx[0])
    decreasing_idx = np.where(np.diff(x) < 0)
    # dec_len = len(decreasing_idx[0])
    # stationary_idx = np.where(np.diff(x) == 0)
    increasing = group_consecutives(increasing_idx)
    decreasing = group_consecutives(decreasing_idx)
    if not (len(increasing) == 1 and len(decreasing) == 1):
        raise ValueError(
            "Values could not be split into two strictly monotonic parts safely."
        )
    if flip_increasing:
        x[increasing_idx[0][0]] *= -1
        x[increasing_idx[0] + 1] *= -1
        return x
    x[decreasing_idx[0]] *= -1
    return x


def group_consecutives(arr, step=1):
    return np.split(arr, np.where(np.diff(arr) != step)[0] + 1)


def from_pat(pat, mod=None):
    if mod is None:
        return {'ifg_names': glob(pat)}
    elif mod == 3:
        files = sorted(glob(pat))
        return {
            'ifg_names': files[::mod],
            'sam_names': files[1::mod],
            'ref_names': files[2::mod]
        }
    elif mod == -1:
        files = sorted(glob(pat))
        return {
            'ifg_names': files[::mod]
        }
    else:
        raise ValueError(
            f"mod = {mod} is ambigous. Consider using your own logic for selecting files."
        )
    


import logging
import warnings

import numpy as np

from pysprint.config import _get_config_value
from pysprint.core.bases.dataset import Dataset
from pysprint.mpl_tools.peak import EditPeak
from pysprint.core.phase import Phase
from pysprint.core._evaluate import min_max_method, is_inside
from pysprint.utils import (
    _maybe_increase_before_cwt,
    _calc_envelope,
)
from pysprint.utils import PySprintWarning


logger = logging.getLogger(__name__)
FORMAT = "[ %(filename)s:%(lineno)s - %(funcName)20s() ] %(message)s"
logging.basicConfig(format=FORMAT)

__all__ = ["MinMaxMethod"]


class MinMaxMethod(Dataset):
    """
    Interface for Minimum-Maximum Method.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.phase = None
        self._is_onesided = False

    def init_edit_session(self, engine="normal", **kwargs):
        """
        Function to initialize peak editing on a plot.
        Right clicks (`d` key later) will delete the closest point,
        left clicks(`i` key later) will add a new point. Just close
        the window when finished. Must be called with interactive
        backend. The best practice is to call this function inside
        `~pysprint.interactive` context manager.

        Parameters
        ----------
        engine : str, optional
            Must be 'cwt', 'normal' or 'slope'.
            Peak detection algorithm to use.
            Default is normal.
        kwargs : dict, optional
            pmax, pmin, threshold, except_around, width
        """
        engines = ("cwt", "normal", "slope")

        if engine not in engines:
            raise ValueError(f"Engine must be in {str(engines)}")

        if engine == "normal":
            pmax = kwargs.pop("pmax", 0.1)
            pmin = kwargs.pop("pmin", 0.1)
            threshold = kwargs.pop("threshold", 0.1)
            except_around = kwargs.pop("except_around", None)
            side = kwargs.pop("side", "both")
            _x, _y, _xx, _yy = self.detect_peak(
                pmax=pmax, pmin=pmin, threshold=threshold, except_around=except_around, side=side
            )

            # just for validation purposes
            _ = kwargs.pop("widths", np.arange(1, 20))
            _ = kwargs.pop("floor_thres", 0.05)

        elif engine == "slope":
            side = kwargs.pop("side", "both")
            self._is_onesided = side != "both"
            x, _, _, _ = self._safe_cast()
            y = np.copy(self.y_norm)
            if _maybe_increase_before_cwt(y):
                y += 2
            _, lp, lloc = _calc_envelope(y, np.arange(len(y)), "l")
            _, up, uloc = _calc_envelope(y, np.arange(len(y)), "u")
            lp -= 2
            up -= 2
            _x, _xx = x[lloc], x[uloc]
            _y, _yy = lp, up

        elif engine == "cwt":
            widths = kwargs.pop("widths", np.arange(1, 20))
            side = kwargs.pop("side", "both")
            floor_thres = kwargs.pop("floor_thres", 0.05)
            _x, _y, _xx, _yy = self.detect_peak_cwt(
                widths=widths, floor_thres=floor_thres, side=side
            )

            # just for validation purposes
            _ = kwargs.pop("pmax", 0.1)
            _ = kwargs.pop("pmin", 0.1)
            _ = kwargs.pop("threshold", 0.1)
            _ = kwargs.pop("except_around", None)

        if side == "both":
            _xm = np.append(_x, _xx)
            _ym = np.append(_y, _yy)
        elif side == "min":
            _xm, _ym = _xx, _yy
        elif side == "max":
            _xm, _ym = _x, _y
        else:
            raise ValueError("Side must be 'both', 'min' or 'max'.")

        if kwargs:
            raise TypeError(f"Invalid argument:{kwargs}")

        try:
            _editpeak = EditPeak(self.x, self.y_norm, _xm, _ym)
        except ValueError:
            _editpeak = EditPeak(self.x, self.y, _xm, _ym)
        # Automatically propagate these points to the mins and maxes.
        # Distribute these points between min and max, just in case
        # the default argrelextrema is definitely not called
        # in `pysprint.core.evaluate.min_max_method`.

        self.xmin = _editpeak.get_dat[0][:len(_editpeak.get_dat[0]) // 2]
        self.xmax = _editpeak.get_dat[0][len(_editpeak.get_dat[0]) // 2:]
        print(f"{len(_editpeak.get_dat[0])} extremal points were recorded.")
        return _editpeak.get_dat[0]  # we should return None

    def calculate(
            self,
            reference_point,
            order,
            SPP_callbacks=None,
            show_graph=False,
            scan=False,
            onesided=False,
    ):
        """
        MinMaxMethod's calculate function.

        Parameters
        ----------
        reference_point : float
            reference point on x axis
        order : int
            Polynomial (and maximum dispersion) order to fit. Must be in [1, 5].
        SPP_callbacks : number, or numeric list-like
            The positions of SPP's on the interferogram. If not given it will check
            if there's any SPP position set on the object.
        show_graph : bool, optional
            Shows a the final graph of the spectral phase and fitted curve.
            Default is False.
        onesided : bool
            Use only minimums or maximums to build the phase. It also works for
            one characteristic point per oscillation period (e.g. zero-crossings).
            Default is False.

        Returns
        -------
        dispersion : array-like
            [GD, GDD, TOD, FOD, QOD, SOD]
        dispersion_std : array-like
            Standard deviations due to uncertainty of the fit.
            They are only calculated if lmfit is installed.
            [GD_std, GDD_std, TOD_std, FOD_std, QOD_std, SOD_std]
        fit_report : str
            lmfit report if installed, else empty string.

        Note
        ----
        Decorated with pprint_disp, so the results are
        immediately printed without explicitly saying so.
        """

        phase = self.build_phase(
            reference_point=reference_point, SPP_callbacks=SPP_callbacks, onesided=onesided
        )

        if is_inside(reference_point, phase.x):
            left_phase = phase.slice(None, reference_point, inplace=False)
            right_phase = phase.slice(reference_point, None, inplace=False)
        else:
            scan = False
            logger.info("Scan is disabled, reference_point is on the border.")

        if scan:
            left_d, left_ds, left_fit_report = left_phase._fit(reference_point, order)
            right_d, right_ds, right_fit_report = right_phase._fit(reference_point, order)

            logger.info(f"left side evaluated to {left_d}, used {len(left_phase.x)} points.")
            logger.info(f"right side evaluated to {right_d}, used {len(right_phase.x)} points.")

            right_d = np.where(np.sign(left_d) != np.sign(right_d), -right_d, right_d)
            right_ds = np.where(np.sign(left_ds) != np.sign(right_ds), -right_ds, right_ds)

            diffs = np.abs(np.trim_zeros(right_d) - np.trim_zeros(left_d)) / np.trim_zeros(left_d)

            thres = _get_config_value("scan_threshold")

            if (np.abs(diffs[~np.isnan(diffs)]) > thres).any():
                dispersion = left_d if len(left_phase.x) >= len(right_phase.x) else right_d
                dispersion_std = left_ds if len(left_phase.x) >= len(right_phase.x) else right_ds
                fit_report = left_fit_report if len(left_phase.x) >= len(right_phase.x) else right_fit_report
                side = "left" if len(left_phase.x) >= len(right_phase.x) else "right"
                logger.info(f"Max relative difference is too high, using {side} side.")

                if side == "left":
                    if show_graph:
                        phase.plot()
                        left_phase.plot(label="used part")
                else:
                    if show_graph:
                        phase.plot()
                        right_phase.plot(label="used part")

            else:
                dispersion = np.mean([left_d, right_d], axis=0)

                dispersion_std = np.mean([left_ds, right_ds], axis=0)

                fit_report = ''.join([left_fit_report, right_fit_report])
                if show_graph:
                    left_phase.plot()
                    right_phase.plot()
        else:
            dispersion, dispersion_std, fit_report = phase._fit(reference_point, order)
            if show_graph:
                phase.plot()

        return dispersion, dispersion_std, fit_report

    def build_phase(self, reference_point, SPP_callbacks=None, onesided=False):
        """
        Build **only the phase** using reference point and SPP positions.

        Parameters
        ----------
        reference_point : float
            The reference point from where the phase building starts.
        SPP_callbacks : number, or numeric list-like
            The positions of SPP's on the interferogram. If not given it will check
            if there's any SPP position set on the object.
        onesided : bool
            If `True`, use only the minimums or maximums to build the phase. It also
            works for one characteristic point per oscillation period (e.g. zero-crossings).
            Default is False.

        Returns
        -------
        phase : pysprint.core.phase.Phase
            The phase object. See its docstring for more info.
        """
        if SPP_callbacks is None and self._positions is not None:
            SPP_callbacks = np.array(self._positions)

        if onesided and not self._is_onesided:
            warnings.warn(
                "Trying to build phase as one-sided, but the detection was two-sided. Use `onesided=False`.",
                PySprintWarning
            )

        if not onesided and self._is_onesided:
            warnings.warn(
                "Trying to build phase as two-sided, but the detection was one-sided. Use `onesided=True`.",
                PySprintWarning
            )

        x, y = min_max_method(
            self.x,
            self.y,
            self.ref,
            self.sam,
            ref_point=reference_point,
            maxx=self.xmax,
            minx=self.xmin,
            SPP_callbacks=SPP_callbacks,
            onesided=onesided,
        )
        self.phase = Phase(x, y)
        return self.phase



"""
The implementation is mostly taken from the pandas library with some simplifications
(e.g. deprecations are left out).
# Copyright (c) Pandas Development Team.
# Distributed under the terms of the BSD 3-Clause "New" or "Revised" License.
"""

import re
import keyword
import tokenize
from collections import namedtuple
from contextlib import ContextDecorator


class ConfigError(ValueError):
    pass


RegisteredConfigValue = namedtuple("RegisteredConfigValue", "key default doc validator")

_registered_config = {}

_global_config = {}


def _select_option(pattern):
    if pattern in _registered_config:
        return [pattern]

    keys = sorted(_registered_config.keys())

    return [k for k in keys if re.search(pattern, k, re.I)]


def _get_single_key(pattern):
    keys = _select_option(pattern)
    if len(keys) == 0:
        raise ConfigError(f"No such key(s): {repr(pattern)}")
    if len(keys) > 1:
        raise ConfigError("Pattern matched multiple keys")
    key = keys[0]
    return key


def _get_config_value(pattern):
    key = _get_single_key(pattern)
    return _global_config[key]


def _set_config_value(*args, **kwargs):
    nargs = len(args)
    if not nargs or nargs % 2 != 0:
        raise ValueError("Must provide an even number of non-keyword arguments")

    if kwargs:
        kwarg = list(kwargs.keys())[0]
        raise TypeError(f'_set_config_value() got an unexpected keyword argument "{kwarg}"')

    for k, v in zip(args[::2], args[1::2]):
        key = _get_single_key(k)

        opt = _registered_config.get(key)
        if opt and opt.validator:
            opt.validator(v)

        _global_config[key] = v


def register_config_value(key, default, doc, validator=None):
    """
    Top level function which sets up the config.
    """
    key = key.lower()

    if key in _registered_config:
        raise ConfigError(f"Option '{key}' has already been registered.")

    if validator:
        validator(default)

    path = key.split(".")

    for k in path:
        if not re.match("^" + tokenize.Name + "$", k):
            raise ValueError(f"{k} is not a valid identifier.")
        if keyword.iskeyword(k):
            raise ValueError(f"{k} is a python keyword.")

    loc = _global_config
    msg = "'{option}' is already an option"

    for i, p in enumerate(path[:-1]):
        if not isinstance(loc, dict):
            raise ConfigError(msg.format(option=".".join(path[:i])))
        if p not in loc:
            loc[p] = {}
        loc = loc[p]

    if not isinstance(loc, dict):
        raise ConfigError(msg.format(option=".".join(path[:-1])))

    # initialize the default config
    loc[path[-1]] = default

    # write metadata
    _registered_config[key] = RegisteredConfigValue(
        key=key, default=default, doc=doc, validator=validator
    )


def _describe_config_value(pattern):

    keys = _select_option(pattern)
    if len(keys) == 0:
        raise ConfigError("No such key(s).")

    out = "\n".join([_build_config_value_description(k) for k in keys])

    return out


def _build_config_value_description(key):
    option = _registered_config.get(key)

    out = f"{key} "

    if option.doc:
        out += "\n".join(option.doc.strip().split("\n"))
    else:
        out += "No description."

    if option:
        out += f"\n value: {_get_config_value(key)} (defaults to {option.default})"
    return out


class setting(ContextDecorator):
    def __init__(self, *args):
        if not (len(args) % 2 == 0 and len(args) >= 2):
            raise ConfigError

        self.ops = list(zip(args[::2], args[1::2]))

    def __enter__(self):
        self.restore = [(pattern, _get_config_value(pattern)) for pattern, val in self.ops]

        for pattern, val in self.ops:
            _set_config_value(pattern, val)

    def __exit__(self, *args):
        if self.restore:
            for pattern, val in self.restore:
                _set_config_value(pattern, val)



import numpy as np
import matplotlib.pyplot as plt

from pysprint.core.bases._dataset_base import _DatasetBase, C_LIGHT
from pysprint.core._generator import generator_freq, generator_wave

__all__ = ["Generator"]


class Generator(metaclass=_DatasetBase):
    """
    Basic dataset generator.
    """
    def __init__(  # TODO : add docstring.
        self,
        start,
        stop,
        center,
        delay=0,
        GD=0,
        GDD=0,
        TOD=0,
        FOD=0,
        QOD=0,
        SOD=0,
        resolution=0.1,
        pulse_width=10,
        normalize=False,
        chirp=0,
    ):
        self.start = start
        self.stop = stop
        self.center = center
        self.delay = delay
        self.GD = GD
        self.GDD = GDD
        self.TOD = TOD
        self.FOD = FOD
        self.QOD = QOD
        self.SOD = SOD
        self.resolution = resolution
        self.pulse_width = pulse_width
        self.chirp = chirp
        self.normalize = normalize
        self.x = np.array([])
        self.y = np.array([])
        self.ref = np.array([])
        self.sam = np.array([])
        self.plotwidget = plt
        self.is_wave = False

    def __str__(self):
        return self.__repr__()

    # TODO: PEP8 that horrible line below
    def __repr__(self):
        return f"Generator({self.start}, {self.stop}, {self.center}, delay={self.delay}, GD={self.GD}, GDD={self.GDD}, TOD={self.TOD}, FOD={self.FOD}, QOD={self.QOD}, SOD={self.SOD}, resolution={self.resolution}, pulse_width={self.pulse_width}, normalize={self.normalize})"

    def _check_norm(self):
        """
        Do the normalization when we can.
        """
        if len(self.ref) != 0:
            self._y = (self.y - self.ref - self.sam) / (
                2 * np.sqrt(self.sam * self.ref)
            )

    def generate(self, force_wavelength=False):
        """
        Intelligently (kind of) decide what domain we generate the dataset on.
        """
        if force_wavelength:
            self.generate_wave()
        else:
            if self.stop < 100:
                self.generate_freq()
            else:
                self.generate_wave()

    def generate_freq(self):
        """
        Use this to generate the spectrogram in ang. frequency domain.
        """
        self.x, self.y, self.ref, self.sam = generator_freq(
            self.start,
            self.stop,
            self.center,
            self.delay,
            self.GD,
            self.GDD,
            self.TOD,
            self.FOD,
            self.QOD,
            self.SOD,
            self.resolution,
            self.pulse_width,
            self.normalize,
            self.chirp,
        )

    def generate_wave(self):
        """
        Use this to generate the spectrogram in wavelength domain.
        """
        self.is_wave = True
        self.x, self.y, self.ref, self.sam = generator_wave(
            self.start,
            self.stop,
            self.center,
            self.delay,
            self.GD,
            self.GDD,
            self.TOD,
            self.FOD,
            self.QOD,
            self.SOD,
            self.resolution,
            self.pulse_width,
            self.normalize,
            self.chirp,
        )

    def GD_lookup(self):
        return self.GD + self.delay

    def show(self):
        """
        Draws the plot of the generated data.
        """
        self._check_norm()
        if np.iscomplexobj(self.y):
            self.plotwidget.plot(self.x, np.abs(self.y))
        else:
            try:
                self.plotwidget.plot(self.x, self._y, "r")
            except Exception:  # TODO: better exception case
                self.plotwidget.plot(self.x, self.y, "r")
        self.plotwidget.grid()
        self.plotwidget.show(block=True)

    # TODO: rewrite this in a more intelligent manner, this is deprecated
    def save(self, name, path=None):
        """
        Saves the generated dataset with numpy.savetxt.

        Parameters
        ----------

        name: string
            Name of the output file. You shouldn't include the .txt at the end.
        path: string, default is None
            You can also specify the save path.
            e.g path='C:/examplefolder'
            """
        if not name.endswith(".txt"):
            name += ".txt"
        if path is None:
            np.savetxt(
                f"{name}",
                np.column_stack((self.x, self.y, self.ref, self.sam)),
                delimiter=",",
            )
            print(f"Successfully saved as {name}.")
        else:
            np.savetxt(
                f"{path}/{name}",
                np.column_stack((self.x, self.y, self.ref, self.sam)),
                delimiter=",",
            )
            print(f"Successfully saved as {name}.")

    def _phase(self, j):
        if self.is_wave:
            lam = np.arange(self.start, self.stop + self.resolution, self.resolution)
            omega = (2 * np.pi * C_LIGHT) / lam
            omega0 = (2 * np.pi * C_LIGHT) / self.center
            j = omega - omega0
        else:
            lamend = (2 * np.pi * C_LIGHT) / self.start
            lamstart = (2 * np.pi * C_LIGHT) / self.stop
            lam = np.arange(lamstart, lamend + self.resolution, self.resolution)
            omega = (2 * np.pi * C_LIGHT) / lam
            j = omega - self.center
        return (
            j
            + self.delay * j
            + j * self.GD
            + (self.GDD / 2) * j ** 2
            + (self.TOD / 6) * j ** 3
            + (self.FOD / 24) * j ** 4
            + (self.QOD / 120) * j ** 5
            + (self.SOD / 720) * j ** 6
        )

    def phase_graph(self):
        """
        Plots the spectrogram along with the spectral phase.
        """
        self._check_norm()
        self.fig, self.ax = self.plotwidget.subplots(2, 1, figsize=(8, 7))
        self.plotwidget.subplots_adjust(top=0.95)
        try:
            self.ax[0].plot(self.x, self._y, "r")
        except Exception:  # TODO : handle that too
            self.ax[0].plot(self.x, self.y, "r")
        try:
            self.ax[1].plot(self.x, self._phase(self.x))
        except Exception:
            raise ValueError("The spectrum is not generated yet.")

        self.ax[0].set(xlabel="Frequency/Wavelength", ylabel="Intensity")
        self.ax[1].set(xlabel="Frequency/Wavelength", ylabel=r"$\Phi $[rad]")
        self.ax[0].grid()
        self.ax[1].grid()
        self.plotwidget.show(block=True)

    @property
    def data(self):
        """
        Unpacks the generated data.
        If arms are given it returns x, y, reference_y, sample_y
        Else returns x, y
        """
        if len(self.ref) == 0:
            return self.x, self.y
        return self.x, self.y, self.ref, self.sam

    # def pulse_shape(self):
    #     """
    #     Plot the shape of the pulse in the time domain.
    #     """
    #     if not self.normalize:
    #         raise ValueError("Must set normalize=True.")
    #     x_spaced = np.linspace(
    #         self.x[0], self.x[-1], len(self.x)
    #     )
    #     y_phase = self._phase(x_spaced)
    #     timestep = np.diff(x_spaced)[0]
    #     E_field = np.sqrt(self.sam) * np.exp(-1j * y_phase)
    #     E_pulse = np.abs(np.fft.ifft(E_field)) ** 2
    #     x_axis = np.fft.fftfreq(len(self.x), d=timestep / (2 * np.pi))
    #     self.plotwidget.fill_between(x_axis, E_pulse, np.zeros_like(E_pulse), color="red")
    #     self.plotwidget.show(block=True)
    #     return x_axis, E_pulse



import numpy as np
import matplotlib
import matplotlib.pyplot as plt

# ---------------------------MONKEY PATCH-------------------------------------
import matplotlib.cbook as cbook
import matplotlib.widgets as widgets
from matplotlib.backend_tools import ToolBase  # lgtm [py/unused-import]
from matplotlib.backend_managers import ToolManager  # lgtm [py/unused-import]


def tm_init(self, figure=None):
    self._key_press_handler_id = None

    self._tools = {}
    self._keys = {}
    self._toggled = {}
    self._callbacks = cbook.CallbackRegistry()

    # to process keypress event
    self.keypresslock = widgets.LockDraw()
    self.messagelock = widgets.LockDraw()

    self._figure = None
    self.set_figure(figure)


matplotlib.backend_managers.ToolManager.__init__ = tm_init


def tb_init(self, toolmanager, name):
    self._name = name
    self._toolmanager = toolmanager
    self._figure = None


matplotlib.backend_tools.ToolBase.__init__ = tb_init

# ----------------------------------------------------------------------------

from matplotlib.backend_tools import ToolToggleBase
from pysprint.utils import _get_closest


class SelectButton(ToolToggleBase):
    """
    Toggle button on matplotlib toolbar.
    """

    description = "Enable click records"
    default_toggled = True
    default_keymap = "t"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class EditPeak:
    """
    This class helps to record and delete peaks from a dataset.
    Right clicks will delete the closest extremal point found
    on the graph, left clicks will add a new point.
    Edits can be saved by just closing the matplotlib window.
    Returns the x coordinates of the selected points.
    Note that this class shouldn't be explicitly called by the user.
    """
    def __init__(self, x, y, x_extremal=None, y_extremal=None):

        # This is here because we do not want other figures
        # to be affected by this.
        plt.ion()
        matplotlib.rcParams["toolbar"] = "toolmanager"

        self.figure, self.ax = plt.subplots()
        self.cid = None
        self.x = x
        self.y = y
        plt.plot(self.x, self.y, "r")
        self.x_extremal = x_extremal
        self.y_extremal = y_extremal
        self.press()
        if self.x_extremal is not None and self.y_extremal is not None:
            if not len(self.x_extremal) == len(self.y_extremal):
                raise ValueError("Data shapes are different")
        else:
            self.x_extremal, self.y_extremal = [], []
        (self.lins,) = self.ax.plot(
            self.x_extremal, self.y_extremal, "ko", markersize=8, zorder=99, alpha=0.5
        )
        self.ax.grid(alpha=0.7)
        # adding the button to navigation toolbar
        tm = self.figure.canvas.manager.toolmanager
        tm.add_tool("Toggle recording", SelectButton)
        self.figure.canvas.manager.toolbar.add_tool(
            tm.get_tool("Toggle recording"), "toolgroup"
        )
        self.my_select_button = tm.get_tool("Toggle recording")
        plt.sca(self.ax)
        plt.show(block=True)

    def on_clicked(self, event):
        """
        Function to record and discard points on plot.
        This function is kind of expensive, but we must work in pixel
        coordinates, because otherwise when the axes are differently scaled the
        result will be catastrophicly unintuitive to work with.
        """
        ix, iy = event.xdata, event.ydata

        # change the cursor's position to pixels
        xy_pixels = self.ax.transData.transform(np.vstack([ix, iy]).T)
        ix, iy = xy_pixels.T

        # compute the current pixel positions of the curve
        xy_pixels = self.ax.transData.transform(np.vstack([self.x, self.y]).T)
        x_pix, y_pix = xy_pixels.T

        if event.inaxes is None:
            return
        if self.my_select_button.toggled:
            if event.key == "d":
                # compute extremals' pixel coords
                extr_pixels = self.ax.transData.transform(np.vstack([self.x_extremal, self.y_extremal]).T)
                extr_x, extr_y = extr_pixels.T

                # get the closest extremal to the cursor in pixels
                ix, iy, idx = _get_closest(ix, iy, extr_x, extr_y)

                # delete the corresponding index in original array
                self.x_extremal = np.delete(self.x_extremal, idx)
                self.y_extremal = np.delete(self.y_extremal, idx)

            elif event.key == "i":
                # compute the closest point to the cursor in pixels
                ix, iy, idx = _get_closest(ix, iy, x_pix, y_pix)

                # change back to data coords and append
                pixels = self.ax.transData.inverted().transform(np.vstack([ix, iy]).T)
                x_data, y_data = pixels.T

                self.x_extremal = np.append(self.x_extremal, x_data)
                self.y_extremal = np.append(self.y_extremal, y_data)

            self.lins.set_data(self.x_extremal, self.y_extremal)
            plt.draw()
            return

    def press(self):
        """Usual function to connect matplotlib.."""
        self.cid = self.figure.canvas.mpl_connect("key_press_event", self.on_clicked)

    def release(self):
        """ On release functionality. It's never called but we will
        need this later on.."""
        self.figure.canvas.mpl_disconnect(self.cid)

    @property
    def get_dat(self):
        """ Returns the the selected points."""
        return self.lins.get_data()



from math import factorial

import numpy as np

from pysprint.utils import pad_with_trailing_zeros


__all__ = [
    "transform_cf_params_to_dispersion",
    "transform_lmfit_params_to_dispersion",
]


def transform_cf_params_to_dispersion(popt, drop_first=True, dof=0):
    if drop_first:
        popt = popt[1:]

    _disp = popt

    for idx in range(len(popt)):
        _disp[idx] = popt[idx] * factorial(idx + dof)

    _disp = pad_with_trailing_zeros(_disp, 6)
    return _disp, np.zeros(6)


def transform_lmfit_params_to_dispersion(popt, popt_std, drop_first=True, dof=1):
    if drop_first:
        popt, popt_std = popt[1:], popt_std[1:]

    _disp, _disp_std = popt, popt_std

    for idx in range(len(popt)):
        _disp[idx] = popt[idx] * factorial(idx + dof)
        if popt_std[idx] is not None:
            _disp_std[idx] = popt_std[idx] * factorial(idx + dof)
        else:
            _disp_std[idx] = 0
    _disp = pad_with_trailing_zeros(popt, 6)
    _disp_std = pad_with_trailing_zeros(popt_std, 6)
    return _disp, _disp_std



import numpy as np


__all__ = [
    "cos_fit1",
    "cos_fit2",
    "cos_fit3",
    "cos_fit4",
    "cos_fit5",
    "cos_fit6",
    "poly1",
    "poly2",
    "poly3",
    "poly4",
    "poly5",
    "poly6",
    "_fit_config",
    "_cosfit_config",
]


def poly6(x, b0, b1, b2, b3, b4, b5, b6):
    """
    Taylor polynomial for fit
    b1 = GD
    b2 = GDD / 2
    b3 = TOD / 6
    b4 = FOD / 24
    b5 = QOD / 120
    b6 = SOD / 720
    """
    return (
        b0
        + b1 * x
        + b2 * x ** 2
        + b3 * x ** 3
        + b4 * x ** 4
        + b5 * x ** 5
        + b6 * x ** 6
    )


def poly5(x, b0, b1, b2, b3, b4, b5):
    """
    Taylor polynomial for fit
    b1 = GD
    b2 = GDD / 2
    b3 = TOD / 6
    b4 = FOD / 24
    b5 = QOD / 120
    """
    return b0 + b1 * x + b2 * x ** 2 + b3 * x ** 3 + b4 * x ** 4 + b5 * x ** 5


def poly4(x, b0, b1, b2, b3, b4):
    """
    Taylor polynomial for fit
    b1 = GD
    b2 = GDD / 2
    b3 = TOD / 6
    b4 = FOD / 24
    """
    return b0 + b1 * x + b2 * x ** 2 + b3 * x ** 3 + b4 * x ** 4


def poly3(x, b0, b1, b2, b3):
    """
    Taylor polynomial for fit
    b1 = GD
    b2 = GDD / 2
    b3 = TOD / 6

    """
    return b0 + b1 * x + b2 * x ** 2 + b3 * x ** 3


def poly2(x, b0, b1, b2):
    """
    Taylor polynomial for fit
    b1 = GD
    b2 = GDD / 2
    """
    return b0 + b1 * x + b2 * x ** 2


def poly1(x, b0, b1):
    """
    Taylor polynomial for fit
    b1 = GD
    """
    return b0 + b1 * x


def cos_fit1(x, c0, c1, b0, b1):
    return c0 + c1 * np.cos(poly1(x, b0, b1))


def cos_fit2(x, c0, c1, b0, b1, b2):
    return c0 + c1 * np.cos(poly2(x, b0, b1, b2))


def cos_fit3(x, c0, c1, b0, b1, b2, b3):
    return c0 + c1 * np.cos(poly3(x, b0, b1, b2, b3))


def cos_fit4(x, c0, c1, b0, b1, b2, b3, b4):
    return c0 + c1 * np.cos(poly4(x, b0, b1, b2, b3, b4))


def cos_fit5(x, c0, c1, b0, b1, b2, b3, b4, b5):
    return c0 + c1 * np.cos(poly5(x, b0, b1, b2, b3, b4, b5))


def cos_fit6(x, c0, c1, b0, b1, b2, b3, b4, b5, b6):
    return c0 + c1 * np.cos(poly6(x, b0, b1, b2, b3, b4, b5, b6))


_fit_config = {1: poly1, 2: poly2, 3: poly3, 4: poly4, 5: poly5, 6: poly6}


_cosfit_config = {
    1: cos_fit1,
    2: cos_fit2,
    3: cos_fit3,
    4: cos_fit4,
    5: cos_fit5,
    6: cos_fit6
}



import os
import unittest
from unittest.mock import patch

import pytest
import numpy as np
import matplotlib.pyplot as plt

from pysprint import MinMaxMethod


class TestEvaluate(unittest.TestCase):
    def setUp(self):
        self.x = np.arange(1, 1000, 1)
        self.y = np.sin(self.x)

    def tearDown(self):
        pass

    @pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure Pipelines fails this.")
    @patch("matplotlib.pyplot.show")
    def test_edit_session(self, mock_show):
        ifg = MinMaxMethod(self.x, self.y)
        ifg.init_edit_session(engine="cwt")
        ifg.init_edit_session(engine="slope")
        ifg.init_edit_session(engine="normal")
        plt.close("all")
        mock_show.assert_called()

    def test_edit_session2(self):
        ifg = MinMaxMethod(self.x, self.y)
        with self.assertRaises(ValueError):
            ifg.init_edit_session(engine="dfssdfasdf")

    def test_edit_session3(self):
        ifg = MinMaxMethod(self.x, self.y)
        with self.assertRaises(TypeError):
            ifg.init_edit_session(engine="normal", invalidkwarg=3)

    @pytest.mark.skipif("TF_BUILD" in os.environ, reason="Fails on azure.")
    @patch("matplotlib.pyplot.show")
    def test_edit_session4(self, mock_show):
        ifg = MinMaxMethod(self.x, self.y)
        ifg.init_edit_session()
        plt.close("all")
        mock_show.assert_called()


def test_phase_build():
    m = MinMaxMethod(np.arange(10), np.arange(10))
    m.xmin = np.arange(0, 10)[::2]
    m.xmax = np.arange(0, 10)[1::2]

    m.build_phase(reference_point=0)

    np.testing.assert_array_equal(np.arange(10), m.phase.x)
    np.testing.assert_array_almost_equal(
        np.array(
            [
                -3.141593,
                -3.141593,
                -6.283185,
                -9.424778,
                -12.566371,
                -15.707963,
                -18.849556,
                -21.991149,
                -25.132741,
                -28.274334
            ]
        ),
        m.phase.y
    )


def test_calculate():
    m = MinMaxMethod(np.arange(10), np.arange(10))
    m.xmin = np.arange(0, 10)[::2]
    m.xmax = np.arange(0, 10)[1::2]

    d, ds, _ = m.calculate(reference_point=0, order=1)

    np.testing.assert_array_almost_equal(d, [-2.970233])

    np.testing.assert_array_equal(np.arange(10), m.phase.x)
    np.testing.assert_array_almost_equal(
        np.array(
            [
                -3.141593,
                -3.141593,
                -6.283185,
                -9.424778,
                -12.566371,
                -15.707963,
                -18.849556,
                -21.991149,
                -25.132741,
                -28.274334
            ]
        ),
        m.phase.y
    )


if __name__ == "__main__":
    unittest.main()



import pytest
import numpy as np

from pysprint.core.bases import Dataset
from pysprint.core.methods import FFTMethod
from pysprint.core.methods.generator import Generator


@pytest.fixture()
def setup_ifg():
    x = np.linspace(-1, 8, 1000)
    y = np.sin(x) * np.cos(x)
    noarm = Dataset(x, y)
    arm = Dataset(x, y, y - 1, y - 2)
    return noarm, arm


def test_axis(setup_ifg):
    a, _ = setup_ifg
    with pytest.raises(ValueError):
        a.transform(None, axis="dsa")


def test_apply_selffunc_noarm(setup_ifg):
    a, _ = setup_ifg
    before = a.x
    a.transform("chdomain", axis=0)
    a.transform("chdomain", axis=0)
    after = a.x

    np.testing.assert_array_almost_equal(before, after)


def test_apply_selffunc_special_axis_arg():

    x = np.linspace(-1, 8, 1000)
    y = np.sin(x) * np.cos(x)
    f = FFTMethod(x, y)
    before = f.x
    f.transform("shift", axis=0)
    f.transform("shift", axis=0)
    after = f.x
    np.testing.assert_array_almost_equal(before, after)


def test_apply_selffunc_arm(setup_ifg):
    _, a = setup_ifg
    before = a.x
    a.transform("chdomain", axis=0)
    a.transform("chdomain", axis=0)
    after = a.x

    np.testing.assert_array_almost_equal(before, after)


def test_apply_np_ufunc_noarm(setup_ifg):
    a, _ = setup_ifg
    before = a.y_norm
    a.transform(np.sqrt, axis=1)
    after = a.y_norm

    np.testing.assert_array_almost_equal(np.sqrt(before), after)


def test_apply_functype(setup_ifg):
    a, _ = setup_ifg

    def f(y, n, k=3):
        if n < y < k:
            return 1
        else:
            return 0

    a.transform(f, args=(0,), kwargs={"k": 0.5}, axis=1)
    assert np.all(a.y[np.where(a.y > 0)][0] == 1)
    assert np.all(a.y[np.where(a.y < 0.5)][0] == 0)


def test_apply_functype2(setup_ifg):
    _, a = setup_ifg

    def f(y, n, k=3):
        if n < y < k:
            return 1
        else:
            return 0

    a.transform(f, args=(0,), kwargs={"k": 0.5}, axis=1)
    assert np.all(a.y[np.where(a.y > 0)][0] == 1)
    assert np.all(a.y[np.where(a.y < 0.5)][0] == 0)


def test_validation(setup_ifg):
    a, _ = setup_ifg

    def splitup(x):
        return -2
    with pytest.raises(ValueError):
        a.transform(splitup, axis=0)


def test_validation2(setup_ifg):
    a, _ = setup_ifg

    @np.vectorize
    def csum(x):
        return np.reshape(x, newshape=(2, -1))
    a.transform(csum, axis=0)



import unittest
import importlib

import pytest
import numpy as np

from pysprint.utils import (
    _handle_input,
    pprint_disp,
    _fourier_interpolate,
    pad_with_trailing_zeros,
    find_nearest,
    measurement,
)


try:
    importlib.import_module("IPython")
    SKIP_PPRINT = True
except ModuleNotFoundError:
    SKIP_PPRINT = False


class TestMisc(unittest.TestCase):
    def setUp(self):
        self.x = np.arange(100)
        self.y = np.sin(self.x)
        self.ref = np.ones_like(self.x)
        self.sam = np.ones_like(self.x)

        self._y = (self.y - self.sam - self.ref) / (2 * np.sqrt(self.sam * self.ref))

    def test_input_handling1(self):
        x, y = _handle_input(self.x, self.y, [], [])
        np.testing.assert_array_equal(x, self.x)
        np.testing.assert_array_equal(y, self.y)

    def test_input_handling2(self):
        x, y = _handle_input(self.x, self.y, self.ref, self.sam)
        np.testing.assert_array_equal(x, self.x)
        np.testing.assert_array_equal(y, self._y)

    def test_input_handling3(self):
        x, y = _handle_input(self.x, self.y, self.ref, [])
        np.testing.assert_array_equal(x, self.x)
        np.testing.assert_array_equal(y, self.y)

    def test_input_handling4(self):
        with self.assertRaises(ValueError):
            _handle_input(self.x, [], [], [])

    def test_input_handling5(self):
        with self.assertRaises(ValueError):
            _handle_input([], self.y, [], [])

    @pytest.mark.skipif(
        SKIP_PPRINT, reason="IPython.display treats this another way"
    )
    def test_pprint(self):
        """source : https://stackoverflow.com/a/4220278/11751294"""
        import sys
        from io import StringIO

        saved_stdout = sys.stdout
        try:
            out = StringIO()
            sys.stdout = out

            @pprint_disp
            def calculation():
                return [0, 1], [0, 1], ""

            calculation()
            output = out.getvalue().strip()
            assert (
                output
                == """GD = 0.00000 ± 0.00000 fs^1
GDD = 1.00000 ± 1.00000 fs^2"""
            )
        finally:
            sys.stdout = saved_stdout

    def test_fourier_interpol(self):
        x, _ = _fourier_interpolate(np.geomspace(10, 1000), np.geomspace(10, 1000))
        np.testing.assert_array_equal(x, np.linspace(10, 1000))

    def test_padding(self):
        arr = np.array([1, 5, 6])
        res = pad_with_trailing_zeros(arr, shape=8)
        assert len(res) == 8

    def test_find_nearest(self):
        x = np.array([1, 2, 4, 9])
        el, idx = find_nearest(x, 10)
        assert el == 9
        assert idx == 3

    def test_measurement(self):
        a = np.array([123.783, 121.846, 122.248, 125.139, 122.569])
        mean, interval = measurement(a, 0.99, silent=True)
        assert mean == 123.117
        assert interval == (120.35397798230359, 125.88002201769642)


if __name__ == "__main__":
    unittest.main()



import sys
import argparse
import pathlib
import datetime

from jinja2 import Environment, PackageLoader

NAME_MAPPING = {
    "cff": 'CosFitMethod',
    "mm": 'MinMaxMethod',
    "spp": 'SPPMethod',
    "fft": "FFTMethod",
    "wft": "WFTMethod"
}


def render(methodname):
    template_env = Environment(loader=PackageLoader('pysprint', 'templates'))
    TEMPLATE_FILE = "method_template.py_t"
    template = template_env.get_template(TEMPLATE_FILE)
    BODY_FILE = methodname + "_body.py_t"
    body = template_env.get_template(BODY_FILE)
    t = template.render(
        methodname=NAME_MAPPING[methodname],
        body=body.render(),
        date=datetime.datetime.now()
    )
    if methodname == "spp":
        SPP_TEMPLATE = "spp_body.py_t"
        template = template_env.get_template(SPP_TEMPLATE)
        t = template.render(
            methodname=NAME_MAPPING[methodname],
            date=datetime.datetime.now()
        )
    path = pathlib.Path(".")
    filename = methodname + "-" + datetime.datetime.today().strftime('%Y-%m-%d-%H-%M') + ".py"
    with open(path / filename, "w") as output_file:
        output_file.write(t)
    print(f"Created template {filename} at {path.absolute()}")


def get_parser():
    parser = argparse.ArgumentParser(
        description=("""
Generate sample evaluation files for different methods.
        """)
    )
    parser.add_argument('-T', '--template', choices=NAME_MAPPING.keys())
    return parser


def main(argv=sys.argv[1:]):
    parser = get_parser()
    try:
        args = parser.parse_args(argv)
    except SystemExit as err:
        return err.code
    render(args.template)
    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



import numpy as np
import pytest

from pysprint.core.bases.algorithms import longest_common_subsequence


def test_lcs_bases():
    x = np.arange(20)
    a, b, c, d = longest_common_subsequence(x, x, x, x)
    np.testing.assert_array_equal(a, x)
    np.testing.assert_array_equal(b, x)
    np.testing.assert_array_equal(c, x)
    np.testing.assert_array_equal(d, x)


def test_lcs_nonbase_fail():
    x = np.arange(20)
    y = x + 0.1
    with pytest.raises(ValueError):
        longest_common_subsequence(x, x, y, y, tol=0.001)


def test_lcs_nonbase():
    x = np.arange(20)
    y = x + 0.1
    a, b, c, d = longest_common_subsequence(x, x, y, y)
    np.testing.assert_array_equal(a, x)
    np.testing.assert_array_equal(b, x)
    np.testing.assert_array_equal(c, y)
    np.testing.assert_array_equal(d, y)



"""
Methods for manipulating the loaded data
"""
import numpy as np
from scipy.signal import (
    find_peaks,
    savgol_filter,
    gaussian,  # TODO : this is deprecated
    convolve,
    find_peaks_cwt,
)
from scipy.interpolate import interp1d
from pysprint.utils import (
    find_nearest,
    _handle_input,
    _between,
    _maybe_increase_before_cwt,
)

# TODO : all of these methods should be rewritten, they are poor quality.


def cwt(x, y, ref, sam, widths, floor_thres=0.1):
    x, y = _handle_input(x, y, ref, sam)
    idx = find_peaks_cwt(y, widths=widths)
    if _maybe_increase_before_cwt(y, tolerance=floor_thres):
        y += 2
    y_rec = 1 / y
    idx2 = find_peaks_cwt(y_rec, widths=widths)
    return x[idx], y[idx] - 2, x[idx2], y[idx2] - 2


def savgol(x, y, ref, sam, window=101, order=3):
    x, y = _handle_input(x, y, ref, sam)
    xint, yint = interpolate_data(x, y, [], [])
    if window > order:
        try:
            if window % 2 == 1:
                fil = savgol_filter(yint, window_length=window, polyorder=order)
                return xint, fil
            else:
                fil = savgol_filter(yint, window_length=window + 1, polyorder=order)
                return xint, fil
        except Exception as e:
            print(e)
    else:
        raise ValueError("Order must be lower than window length.")


def find_peak(x, y, ref, sam, pro_max=1, pro_min=1, threshold=0.1, except_around=None):
    if except_around is not None and len(except_around) != 2:
        raise ValueError("Invalid except_around arg. Try [start, stop].")
    if except_around is not None:
        try:
            float(except_around[0])
            float(except_around[1])
        except ValueError:
            raise ValueError(
                "Invalid except_around arg. Only numeric values are allowed."
            )
    x, y = _handle_input(x, y, ref, sam)
    max_indexes, _ = find_peaks(y, prominence=pro_max)
    y_rec = 1 / y
    min_indexes, _ = find_peaks(y_rec, prominence=pro_min)
    min_idx = []
    max_idx = []
    for idx in max_indexes:
        if _between(x[idx], except_around) or np.abs(y[idx]) > threshold:
            max_idx.append(idx)
    for idx in min_indexes:
        if _between(x[idx], except_around) or np.abs(y[idx]) > threshold:
            min_idx.append(idx)

    if len(x[max_idx]) != len(y[max_idx]) or len(x[min_idx]) != len(y[min_idx]):
        raise ValueError("Something went wrong, try to cut the edges of data.")

    return x[max_idx], y[max_idx], x[min_idx], y[min_idx]


def convolution(x, y, ref, sam, win_len, standev=200):
    x, y = _handle_input(x, y, ref, sam)
    if win_len < 0 or win_len > len(x):
        raise ValueError("Window length must be 0 < window_length < len(x)")

    xint, yint = interpolate_data(x, y, [], [])
    window = gaussian(win_len, std=standev)
    smoothed = convolve(yint, window / window.sum(), mode="same")
    return xint, smoothed


def interpolate_data(x, y, ref, sam):
    x, y = _handle_input(x, y, ref, sam)
    xint = np.linspace(x[0], x[-1], len(x))
    intp = interp1d(x, y, kind="linear")
    yint = intp(xint)
    return xint, yint


def cut_data(x, y, ref, sam, start=None, stop=None):
    x, y = _handle_input(x, y, ref, sam)
    if start is None:
        start = np.min(x)
    if stop is None:
        stop = np.max(x)
    if start < stop:
        low_item, _ = find_nearest(x, start)
        high_item, _ = find_nearest(x, stop)
        mask = np.where((x >= low_item) & (x <= high_item))
        return x[mask], y[mask]
    elif stop < start:
        raise ValueError("Start must not exceed stop value.")
    else:
        return np.array([]), np.array([])



import os
from unittest import mock

import pytest
import numpy as np
from matplotlib import rcParams

from pysprint.mpl_tools.peak import EditPeak
from pysprint.mpl_tools.spp_editor import SPPEditor
from pysprint.mpl_tools.normalize import DraggableEnvelope


rcParams['figure.max_open_warning'] = 30


def mock_event(xdata, ydata, button, key, fig, canvas, inaxes=True):
    event = mock.Mock()
    event.button = button
    event.key = key
    event.xdata, event.ydata = xdata, ydata
    event.x, event.y = xdata, ydata
    event.inaxes = inaxes
    event.fig = fig
    event.canvas = canvas
    event.guiEvent = None
    event.name = 'MockEvent'

    return event


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_insert(mock_show):
    x, y = np.arange(100), np.arange(100)
    xx, yy = np.array([4, 5]), np.array([4, 5])
    obj = EditPeak(x, y, x_extremal=xx, y_extremal=yy)
    mck = mock_event(xdata=50, ydata=50, button="i", key="i", fig=obj.figure, canvas=obj.figure.canvas, inaxes=True)
    obj.on_clicked(event=mck)
    a, b = obj.get_dat
    np.testing.assert_array_almost_equal(a, np.array([4, 5, 50]), decimal=5)
    np.testing.assert_array_almost_equal(b, np.array([4, 5, 50]), decimal=5)
    mock_show.assert_called()


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_delete(mock_show):
    x, y = np.arange(100), np.arange(100)
    xx, yy = np.array([4, 5]), np.array([4, 5])
    obj = EditPeak(x, y, x_extremal=xx, y_extremal=yy)
    mck = mock_event(xdata=50, ydata=50, button="d", key="d", fig=obj.figure, canvas=obj.figure.canvas, inaxes=True)
    obj.on_clicked(event=mck)
    a, b = obj.get_dat
    np.testing.assert_array_equal(a, np.array([4]))
    np.testing.assert_array_equal(b, np.array([4]))
    mock_show.assert_called()


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_not_inaxes(mock_show):
    x, y = np.arange(100), np.arange(100)
    xx, yy = np.array([4, 5]), np.array([4, 5])
    obj = EditPeak(x, y, x_extremal=xx, y_extremal=yy)
    mck = mock_event(xdata=50, ydata=50, button="d", key="d", fig=obj.figure, canvas=obj.figure.canvas, inaxes=None)
    obj.on_clicked(event=mck)
    a, b = obj.get_dat
    np.testing.assert_array_equal(a, xx)
    np.testing.assert_array_equal(b, yy)
    mock_show.assert_called()


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_lock(mock_show):
    x, y = np.arange(100), np.arange(100)
    xx, yy = np.array([4, 5]), np.array([4, 5])
    obj = EditPeak(x, y, x_extremal=xx, y_extremal=yy)
    obj.my_select_button.disable()
    mck = mock_event(xdata=50, ydata=50, button="d", key="d", fig=obj.figure, canvas=obj.figure.canvas, inaxes=None)
    obj.on_clicked(event=mck)
    a, b = obj.get_dat
    np.testing.assert_array_equal(a, xx)
    np.testing.assert_array_equal(b, yy)
    mock_show.assert_called()


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_lock2(mock_show):
    x, y = np.arange(100), np.arange(100)
    xx, yy = np.array([4, 5]), np.array([4, 5])
    obj = EditPeak(x, y, x_extremal=xx, y_extremal=yy)
    obj.release()
    mck = mock_event(xdata=50, ydata=50, button="d", key="d", fig=obj.figure, canvas=obj.figure.canvas, inaxes=None)
    obj.on_clicked(event=mck)
    a, b = obj.get_dat
    np.testing.assert_array_equal(a, xx)
    np.testing.assert_array_equal(b, yy)
    mock_show.assert_called()


def test_inconsistent_length():
    with pytest.raises(ValueError):
        EditPeak(range(500), range(500), x_extremal=[1], y_extremal=[2, 3])


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_normalize_btn_release(mock_show):
    x = np.linspace(0, 6, 10000)
    y = np.cos(x)
    obj = DraggableEnvelope(x, y, mode="u")
    assert obj._ind is None
    y_transform = obj.get_data()
    np.testing.assert_allclose(y_transform[100:9900], np.ones(9800), atol=1, rtol=1)


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_normalize_keypress_cb(mock_show):
    x = np.linspace(0, 6, 1000)
    y = np.cos(x)
    obj = DraggableEnvelope(x, y, mode="u")
    mck = mock_event(xdata=50, ydata=50, button="d", key="d", fig=obj.fig, canvas=obj.fig.canvas, inaxes=True)
    obj.key_press_callback(event=mck)


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_normalize_keypress_cb2(mock_show):
    x = np.linspace(0, 6, 1000)
    y = np.cos(x)
    obj = DraggableEnvelope(x, y, mode="l")
    mck = mock_event(xdata=50, ydata=50, button="i", key="i", fig=obj.fig, canvas=obj.fig.canvas, inaxes=True)
    obj.key_press_callback(event=mck)
    mock_show.assert_called()


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_normalize_keypress_cb3(mock_show):
    x = np.linspace(0, 6, 1000)
    y = np.cos(x)
    with pytest.raises(ValueError):
        DraggableEnvelope(x, y, mode="invalid")


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_normalize_button_press_cb(mock_show):
    x = np.linspace(0, 6, 1000)
    y = np.cos(x)
    obj = DraggableEnvelope(x, y, mode="l")
    mck = mock_event(xdata=50, ydata=50, button="i", key="i", fig=obj.fig, canvas=obj.fig.canvas, inaxes=None)
    obj.button_press_callback(event=mck)
    mock_show.assert_called()


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_normalize_k_press_cb2(mock_show):
    x = np.linspace(0, 6, 1000)
    y = np.cos(x)
    obj = DraggableEnvelope(x, y, mode="l")
    obj.epsilon = 1000000
    obj.x_env = np.array([56, 5])
    obj.y_env = np.array([60, 0.5])
    xy_pixels = obj.ax.transData.transform([5, 0.5])
    xpix, ypix = xy_pixels
    mck = mock_event(xdata=xpix, ydata=ypix, button="d", key="d", fig=obj.fig, canvas=obj.fig.canvas, inaxes=obj.ax)
    obj.key_press_callback(event=mck)
    assert 5 not in obj.x_env
    assert 0.5 not in obj.y_env
    mock_show.assert_called()


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_normalize_keypress_cb4(mock_show):
    x = np.linspace(0, 6, 1000)
    y = np.cos(x)
    obj = DraggableEnvelope(x, y, mode="l")
    mck = mock_event(xdata=2, ydata=0.5, button="i", key="i", fig=obj.fig, canvas=obj.fig.canvas, inaxes=True)
    obj.key_press_callback(event=mck)
    assert 2 in obj.x_env
    assert 0.5 in obj.y_env
    mock_show.assert_called()


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_normalize_motion_notify_cb(mock_show):
    x = np.linspace(0, 6, 1000)
    y = np.cos(x)
    obj = DraggableEnvelope(x, y, mode="l")
    mck = mock_event(xdata=5, ydata=0.5, button="i", key="i", fig=obj.fig, canvas=obj.fig.canvas, inaxes=obj.ax)
    obj.motion_notify_callback(event=mck)
    mock_show.assert_called()


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_sppeditor_submit(mock_show):
    x = np.linspace(0, 6, 1000)
    y = np.cos(x)
    obj = SPPEditor(x, y)
    obj.submit("dsa50.dsa4")
    assert obj.delay == 50.4


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_sppeditor_text_change(mock_show):
    x = np.linspace(0, 6, 1000)
    y = np.cos(x)
    obj = SPPEditor(x, y)
    obj.text_change("dsa50.dsa4")
    assert obj.delay == 50.4


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_sppeditor_btn_release(mock_show):
    x = np.linspace(0, 6, 10000)
    y = np.cos(x)
    obj = SPPEditor(x, y)
    mck = mock_event(xdata=50, ydata=50, button=1, key=1, fig=obj.fig, canvas=obj.fig.canvas, inaxes=None)
    obj.button_release_callback(event=mck)
    assert obj._ind is None


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_sppeditor_get_ind(mock_show):
    x = np.linspace(0, 6, 10000)
    y = np.cos(x)
    obj = SPPEditor(x, y)
    obj.x_pos = np.array([1, 2])
    obj.y_pos = np.array([1, 2])
    mck = mock_event(xdata=1.5, ydata=1.3, button=1, key=1, fig=obj.fig, canvas=obj.fig.canvas, inaxes=obj.ax)
    obj.get_ind_under_point(event=mck)


@pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure fails this.")
@mock.patch("matplotlib.pyplot.show")
def test_sppeditor_keypress_cb(mock_show):
    x = np.linspace(0, 6, 10000)
    y = np.cos(x)
    obj = SPPEditor(x, y)
    obj.x_pos = np.array([1, 2])
    obj.y_pos = np.array([1, 2])
    mck = mock_event(xdata=2, ydata=0, button="i", key="i", fig=obj.fig, canvas=obj.fig.canvas, inaxes=obj.ax)
    obj.key_press_callback(event=mck)
    np.testing.assert_array_equal(obj.x_pos, np.array([1, 2, 2]))
    np.testing.assert_array_equal(obj.y_pos, np.array([1, 2, 0]))


@pytest.mark.skip(reason="Index can't be set..")
@mock.patch("matplotlib.pyplot.show")
def test_sppeditor_keypress_cb2(mock_show):
    x = np.linspace(0, 6, 10000)
    y = np.cos(x)
    obj = SPPEditor(x, y)
    obj.epsilon = 10000
    obj.x_pos = np.array([1, 2, 3])
    obj.y_pos = np.array([1, 2, 3])

    xy_pixels = obj.ax.transData.transform([2, 2])
    xpix, ypix = xy_pixels
    mck = mock_event(xdata=xpix, ydata=ypix, button="d", key="d", fig=obj.fig, canvas=obj.fig.canvas, inaxes=True)
    obj.key_press_callback(event=mck)
    np.testing.assert_array_equal(obj.x_pos, np.array([2]))
    np.testing.assert_array_equal(obj.y_pos, np.array([2]))
    mock_show.assert_called()



"""
Sample generator
"""
import numpy as np

from pysprint.core.bases._dataset_base import C_LIGHT


def _ensure_input(start, stop, center, resolution, pulse_width):
    if start >= stop:
        raise ValueError("start value must be less than stop")
    if center < start or center > stop:
        raise ValueError("center must be _between start and stop")
    if resolution > (stop - start):
        raise ValueError("resolution is too big")
    if pulse_width <= 0:
        raise ValueError("Pulse width is strictly positive.")


def _disp(x, GD=0, GDD=0, TOD=0, FOD=0, QOD=0, SOD=0):
    return (
        x * GD
        + (GDD / 2) * x ** 2
        + (TOD / 6) * x ** 3
        + (FOD / 24) * x ** 4
        + (QOD / 120) * x ** 5
        + (SOD / 720) * x ** 6
    )


def generator_freq(
    start,
    stop,
    center,
    delay,
    GD=0,
    GDD=0,
    TOD=0,
    FOD=0,
    QOD=0,
    SOD=0,
    resolution=0.1,
    pulse_width=10,
    include_arms=False,
    chirp=0,
):
    _ensure_input(start, stop, center, resolution, pulse_width)
    omega0 = center  # unnecessary renaming
    window = (np.sqrt(1 + chirp ** 2) * 8 * np.log(2)) / (pulse_width ** 2)
    lamend = (2 * np.pi * C_LIGHT) / start
    lamstart = (2 * np.pi * C_LIGHT) / stop
    lam = np.arange(lamstart, lamend + resolution, resolution)
    omega = (2 * np.pi * C_LIGHT) / lam
    relom = omega - omega0
    i1 = np.exp(-(relom ** 2) / window)
    i2 = np.exp(-(relom ** 2) / window)
    i = (
        i1
        + i2
        + 2
        * np.cos(
            _disp(relom, GD=GD, GDD=GDD, TOD=TOD, FOD=FOD, QOD=QOD, SOD=SOD) + (omega * delay)
        )
        * np.sqrt(i1 * i2)
    )
    if include_arms:
        return omega, i, i1, i2
    else:
        return omega, i, np.array([]), np.array([])


def generator_wave(
    start,
    stop,
    center,
    delay,
    GD=0,
    GDD=0,
    TOD=0,
    FOD=0,
    QOD=0,
    SOD=0,
    resolution=0.1,
    pulse_width=10,
    include_arms=False,
    chirp=0,
):
    _ensure_input(start, stop, center, resolution, pulse_width)
    omega0 = (2 * np.pi * C_LIGHT) / center
    window = (np.sqrt(1 + chirp ** 2) * 8 * np.log(2)) / (pulse_width ** 2)
    lam = np.arange(start, stop + resolution, resolution)
    omega = (2 * np.pi * C_LIGHT) / lam
    relom = omega - omega0
    i1 = np.exp(-(relom ** 2) / window)
    i2 = np.exp(-(relom ** 2) / window)
    i = (
        i1
        + i2
        + 2
        * np.sqrt(i1 * i2)
        * np.cos(
            _disp(relom, GD=GD, GDD=GDD, TOD=TOD, FOD=FOD, QOD=QOD, SOD=SOD) + (omega * delay)
        )
    )
    if include_arms:
        return lam, i, i1, i2
    else:
        return lam, i, np.array([]), np.array([])



import unittest
from unittest.mock import patch

import numpy as np
import pytest

import pysprint
from pysprint import FFTMethod, Generator
from pysprint.core._fft_tools import (
    find_roi,
    find_center,
    _ensure_window_at_origin,
    predict_fwhm,
)
from pysprint.utils import NotCalculatedException


class TestFFTAuto(unittest.TestCase):
    def test_find_roi(self):
        x, y = find_roi(np.linspace(-10, 10, 10000), np.linspace(-10, 10, 10000))
        np.testing.assert_almost_equal(min(x), 0, decimal=2)
        np.testing.assert_almost_equal(min(y), 0, decimal=2)

    def test_find_center1(self):
        g = Generator(1, 4, 3, 1500)
        g.generate_freq()

        a = FFTMethod(*g.data)
        a.ifft()

        x, y = find_roi(a.x, a.y)

        x_peak, _ = find_center(x, y, n_largest=4)

        assert 1495 < x_peak < 1505

    def test_find_center2(self):
        g = Generator(1, 4, 3, 100)
        g.generate_freq()

        a = FFTMethod(*g.data)
        a.ifft()

        x, y = find_roi(a.x, a.y)

        x_peak, _ = find_center(x, y, n_largest=4)

        assert 95 < x_peak < 105

    def test_find_center3(self):
        g = Generator(1, 4, 3, 7000)
        g.generate_freq()

        a = FFTMethod(*g.data)
        a.ifft()

        x, y = find_roi(a.x, a.y)

        x_peak, _ = find_center(x, y, n_largest=4)

        assert 6950 < x_peak < 7050

    def test_window1(self):
        is_fine, val = _ensure_window_at_origin(0, 200, 4, 10, tol=1e-3)
        assert not is_fine

    def test_window2(self):
        is_fine, val = _ensure_window_at_origin(500, 200, 8, 10, tol=1e-3)
        assert is_fine

    def test_window3(self):
        # test against ambigous overlaps
        is_fine, val = _ensure_window_at_origin(500, 249, 4, 10, tol=1e-3)
        assert is_fine

    def test_window4(self):
        # test against ambigous overlaps
        is_fine, val = _ensure_window_at_origin(500, 251, 4, 10, tol=1e-3)
        assert is_fine

    #  Obviously we need more verbose tests for that later.

    def test_fwhm_detection(self):
        g = Generator(1, 4, 3, 1000)
        g.generate_freq()

        a = FFTMethod(*g.data)
        a.ifft()

        x, y = find_roi(a.x, a.y)

        cent, win_size, order = predict_fwhm(
            x, y, 1000, 10, prefer_high_order=True, tol=1e-3
        )
        assert cent > 1000
        assert win_size > 2000


@pytest.mark.parametrize("printing", [True, False])
def test_autorun(printing):
    g = Generator(1, 4, 3, 1000)
    g.generate()

    f = FFTMethod(*g.data)
    with pytest.raises(NotCalculatedException):
        f.get_pulse_shape_from_array(np.arange(10), np.arange(20))
    with unittest.mock.patch("matplotlib.pyplot.show") as mck:
        phase = f.autorun(enable_printing=printing)
        assert isinstance(phase, pysprint.core.phase.Phase)
        mck.assert_called()
        with pytest.raises(ValueError):
            f.get_pulse_shape_from_array(np.arange(10), np.arange(20))


if __name__ == "__main__":
    unittest.main()



import sys
import time
import threading
import re
from itertools import cycle
from functools import wraps, lru_cache
from copy import copy

import numpy as np

from pysprint.config import _get_config_value
from pysprint.utils.misc import run_from_ipython


__all__ = [
    '_progress',
    'inplacify',
    '_mutually_exclusive_args',
    '_lazy_property',
    'pprint_disp',
]


_inplace_doc = """\n\tinplace : bool, optional
            Whether to apply the operation on the dataset in an "inplace" manner.
            This means if inplace is True it will apply the changes directly on
            the current dataset and returns None. If inplace is False, it will
            leave the current object untouched, but returns a copy of it, and
            the operation will be performed on the copy. It's useful when
            chaining operations on a dataset.\n\n\t"""


def _has_parameter_section(method):
    try:
        return "Parameters" in method.__doc__
    except TypeError:
        return False


def _update_doc(method, doc):
    if _has_parameter_section(method):
        newdoc = _build_doc(method, doc)
        method.__doc__ = newdoc
    else:
        newdoc = "\n\tParameters\n\t----------" + _inplace_doc

        nodoc_head = (f"Docstring automatically created for {method.__name__}. "
                      "Parameter list may not be complete.\n")
        if method.__doc__ is not None:
            method.__doc__ = "".join([method.__doc__, newdoc])
        else:
            method.__doc__ = "".join([nodoc_head, newdoc])
        return


def _build_doc(method, param):
    patt = r"(\w+(?=\s*[-]{4,}[^/]))"  # finding sections
    split_doc = re.split(patt, method.__doc__)
    try:
        target = split_doc.index("Parameters") + 1
    except ValueError:
        return method.__doc__

    split_doc[target] = ''.join([split_doc[target].rstrip(), param])

    return ''.join(filter(None, split_doc))


def inplacify(method):
    """
    Decorator used to allow a class function to be called
    as `inplace` as well. It will invalidate the parent
    object to have **only one** reference to it.
    """
    _update_doc(method, _inplace_doc)

    @wraps(method)
    def wrapper(self, *args, **kwds):
        inplace = kwds.pop("inplace", True)
        if inplace:
            method(self, *args, **kwds)
        else:
            new_ds = method(copy(self), *args, **kwds)

            # trigger a callback to ensure that relevant values
            # aren't dropped
            try:
                new_ds.callback(new_ds, new_ds.parent)
            except (TypeError, ValueError):
                pass
            
            # invalidate parent for the original obj.
            if self.parent is not None:
                self.parent._container.pop(self, None)
                self.parent = None
            return new_ds

    return wrapper


def _progress(func):
    active = threading.Lock()

    def spinning_pbar_printer():
        symbols = ['|', '/', '-', '\\', '\\']
        cursor = cycle(symbols)
        while active.locked():
            sys.stdout.write("\r")
            sys.stdout.write("Working... " + next(cursor))
            sys.stdout.flush()
            time.sleep(0.1)

    def wrapper(*args, **kwargs):
        t = threading.Thread(target=spinning_pbar_printer)
        active.acquire()
        t.start()
        try:
            res = func(*args, **kwargs)
        finally:
            active.release()
        return res

    return wrapper


# https://stackoverflow.com/a/54487188/11751294
def _mutually_exclusive_args(keyword, *keywords):
    """
    Decorator to restrict the user to specify exactly one of the given parameters.
    Often used for std and fwhm for Gaussian windows.
    """
    keywords = (keyword,) + keywords

    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            if sum(k in keywords for k in kwargs) != 1:
                raise TypeError(
                    "You must specify exactly one of {}.".format(" and ".join(keywords))
                )
            return func(*args, **kwargs)

        return inner

    return wrapper


def _lazy_property(f):
    return property(lru_cache()(f))


def pprint_disp(f):
    """
    Pretty print the dispersion results from returned arrays.
    """
    @wraps(f)
    def wrapping(*args, **kwargs):
        disp, disp_std, stri = f(*args, **kwargs)
        labels = ("GD", "GDD", "TOD", "FOD", "QOD", "SOD")
        disp = np.trim_zeros(disp, "b")
        disp_std = disp_std[: len(disp)]
        precision = _get_config_value("precision")
        for i, (label, disp_item, disp_std_item) in enumerate(
                zip(labels, disp, disp_std)
        ):
            if run_from_ipython():
                from IPython.display import display, Math # noqa

                display(
                    Math(f"{label} = {disp_item:.{precision}f} ± {disp_std_item:.{precision}f} fs^{i + 1}")
                )
            else:
                print(f"{label} = {disp_item:.{precision}f} ± {disp_std_item:.{precision}f} fs^{i + 1}")
        return disp, disp_std, stri

    return wrapping



from unittest.mock import patch
import threading

import pytest
import numpy as np

from pysprint.core._optimizer import FitOptimizer
from pysprint import Generator, CosFitMethod


@pytest.mark.skipif(
    not isinstance(threading.current_thread(), threading._MainThread),
    reason="threading breaks matplotlib.."
)
def test_optimizer1():
    x = np.arange(100)
    y = np.sin(x)

    opt = FitOptimizer(x, y, [], [], 50)

    with pytest.raises(ValueError):
        opt.run(r_extend_by=0.1, r_threshold=0.8)


@pytest.mark.skipif(
    not isinstance(threading.current_thread(), threading._MainThread),
    reason="threading breaks matplotlib.."
)
@pytest.mark.parametrize("val", [-10, 500])
def test_optimizer2(val):
    x = np.arange(100)
    y = np.sin(x)

    opt = FitOptimizer(x, y, [], [], 50)
    with pytest.raises(ValueError):
        opt.set_initial_region(val)


@pytest.mark.skipif(
    not isinstance(threading.current_thread(), threading._MainThread),
    reason="threading breaks matplotlib.."
)
def test_optimizer3():
    x = np.arange(100)
    y = np.sin(x)

    opt = FitOptimizer(x, y, [], [], 50)
    opt.set_final_guess(GD=1, GDD=50, TOD=36, FOD=48, QOD=None)
    assert len(opt.rest) == 3
    np.testing.assert_array_almost_equal(opt.rest, [25, 6, 2])
    np.testing.assert_array_almost_equal(opt.user_guess, [25, 6, 2])


# threading somehow breaks it, will be disabled until I figure it out
@pytest.mark.skipif(
    not isinstance(threading.current_thread(), threading._MainThread),
    reason="threading breaks matplotlib.."
)
@pytest.mark.parametrize("GD", [100, -500])
@pytest.mark.parametrize("GDD", [100, -500])
@pytest.mark.parametrize("delay", [0, 400])
def test_optimizer_from_api(delay, GD, GDD):
    g = Generator(1, 3, 2, delay, GD=GD, GDD=GDD, resolution=0.05, normalize=True)
    g.generate_freq()

    cf = CosFitMethod(*g.data)
    cf.guess_GD(GD + delay)
    cf.guess_GDD(GDD)
    with patch.object(FitOptimizer, "update_plot") as patched_obj:
        cf.optimizer(2, order=2, initial_region_ratio=0.01, extend_by=0.01)
        patched_obj.assert_called()


@pytest.mark.skipif(
    not isinstance(threading.current_thread(), threading._MainThread),
    reason="threading breaks matplotlib.."
)
def test_optimizer_from_api2():
    """
    Here we test that granting only GD value will yield correct results.
    """
    g = Generator(
        1,
        3,
        2,
        500,
        GD=400,
        GDD=400,
        TOD=800,
        FOD=7000,
        QOD=70000,
        resolution=0.05,
        normalize=True,
    )
    g.generate_freq()

    cf = CosFitMethod(*g.data)
    cf.guess_GD(900)
    with patch.object(FitOptimizer, "update_plot") as patched_obj:
        res = cf.optimizer(2, order=5, initial_region_ratio=0.01, extend_by=0.01)
        patched_obj.assert_called()
        np.testing.assert_array_almost_equal(res, [900, 400, 800, 7000, 70000])


@pytest.mark.parametrize("delay", list(range(100, 501, 100)))
def test_lookup(delay):
    g = Generator(1, 3, 2, delay, GD=0, GDD=500, TOD=800, normalize=True)
    g.generate_freq()
    cf = CosFitMethod(*g.data)
    cf.GD_lookup(2, engine="normal", silent=True)
    assert delay * 0.95 <= cf.params[3] <= delay * 1.05


def test_lookup2():
    g = Generator(1, 3, 2, 400, GD=0, GDD=500, normalize=False)
    g.generate_freq()
    cf = CosFitMethod(*g.data)
    cf.GD_lookup(200, engine="normal", silent=True)
    assert cf.params[3] == 1



from abc import ABC, abstractmethod
import inspect
import warnings

import numpy as np
import matplotlib.pyplot as plt

from pysprint.utils.decorators import _mutually_exclusive_args
from pysprint.utils.decorators import _lazy_property
from pysprint.core._evaluate import gaussian_window
from pysprint.config import _get_config_value
from pysprint.utils import PySprintWarning


class WindowBase(ABC):

    def __init_subclass__(cls, **kwargs):
        cls._get_params()

    @property
    @abstractmethod
    def y(self):
        pass

    @classmethod
    def _get_params(cls):
        init = getattr(cls.__init__, '..', cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        POPARGS = ('self', 'center', 'x')

        # introspect the constructor arguments
        init_signature = inspect.signature(init)
        for arg in POPARGS:
            if arg not in init_signature.parameters:
                raise ValueError(f"Window class has no `{arg}` parameter in its __init__ signature,"
                                 f" but it must implement all of the following: {', '.join(POPARGS[1:])}.")

        parameters = [p for p in init_signature.parameters.values()
                      if p.name not in POPARGS 
                      and (p.kind == p.VAR_KEYWORD or p.kind == p.VAR_POSITIONAL)]
        for p in parameters:
            raise RuntimeError("Window classes should always "
                           "specify their parameters in the signature"
                           " of their __init__ (no varargs)."
                           f" {cls} with constructor {init_signature} doesn't "
                           f" follow this convention. Suggestion: don't use *args, **kwargs.")

        # Extract argument names excluding POPARGS
        return sorted([p.name for p in parameters])


    def __init__(self, x, center, **kwargs):
        self.x = x
        self.center = center


    def plot(self, ax=None, scalefactor=1, zorder=90, **kwargs):
        """
        Plot the window.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            The axis to plot on. If not given, plot on the last axis.
        scalefactor : float, optional
            Number describing how much a given window should be scaled up ONLY
            for visibility.
        zorder : float, optional
            The drawing order of artists is determined by their zorder attribute, which is
            a floating point number. Artists with higher zorder are drawn on top. You can
            change the order for individual artists by setting their zorder. The default
            value depends on the type of the Artist.
        """
        if ax is None:
            ax = plt
        ax.plot(self.x, self.y * scalefactor, zorder=zorder, **kwargs)


class GaussianWindow(WindowBase):
    """
    Basic class that implements functionality related to Gaussian
    windows with caching the y values.
    """
    @_mutually_exclusive_args('fwhm', 'std')
    def __init__(self, x, center, fwhm=None, std=None, order=2):
        self.x = x
        self.center = center
        self.fwhm = fwhm
        self.order = order
        self.fwhm = fwhm
        self.std = std
        if self.fwhm is None:
            self.fwhm = self.std * 2 * np.log(2) ** (1 / self.order)

    @_lazy_property
    def y(self):
        """
        The y values of the given window. It's a "lazy_property".
        """
        if not np.min(self.x) <= self.center <= np.max(self.x):
            warnings.warn(
                f"Creating window with center {self.center}, which is outside of the dataset's"
                f" range (from {np.min(self.x):.3f} to {np.max(self.x):.3f}).",
                PySprintWarning
            )
        return gaussian_window(self.x, self.center, self.fwhm, self.order)

    def __str__(self):
        precision = _get_config_value("precision")
        return f"Window(center={self.center:.{precision}f}, fwhm={self.fwhm}, order={self.order})"



import sys
import warnings
from inspect import isfunction

import numpy as np
import matplotlib.pyplot as plt

from pysprint.core.methods.fftmethod import FFTMethod
from pysprint.core.phase import Phase
from pysprint.core._fft_tools import find_roi
from pysprint.core._fft_tools import find_center
from pysprint.utils.decorators import inplacify
from pysprint.utils import NotCalculatedException
from pysprint.utils import PySprintWarning
from pysprint.utils.misc import find_nearest
from pysprint.core.window import GaussianWindow
from pysprint.core.window import WindowBase

try:
    from dask import delayed, compute
    from dask.diagnostics import ProgressBar
    CAN_PARALLELIZE = True
except ImportError:
    CAN_PARALLELIZE = False

    def delayed(func=None, *args, **kwargs):
        if isfunction(func):
            return func


class WFTMethod(FFTMethod):
    """Basic interface for Windowed Fourier Transform Method.
    The `window_class` attribuite can be set up for custom windowing.
    """

    def __init__(self, *args, **kwargs):
        self.window_class = kwargs.pop("window_class", GaussianWindow)
        assert issubclass(self.window_class, WindowBase), "window_class must subclass pysprint.core.window.WindowBase"
        super().__init__(*args, **kwargs)
        self.window_seq = {}
        self.found_centers = {}
        self.GD = None
        self.cachedlen = 0
        self.X_cont = np.array([])
        self.Y_cont = np.array([])
        self.Z_cont = np.array([])
        self.fastmath = True
        self.errorcounter = 0

    @inplacify
    def add_window(self, center, **kwargs):
        """
        Add a Gaussian window to the interferogram.

        Parameters
        ----------
        center : float
            The center of the Gaussian window.
        kwargs : dict
            Keyword arguments to pass to the `window_class`.
        """
        window = self.window_class(self.x, center=center, **kwargs)
        self.window_seq[center] = window
        return self

    @property
    def windows(self):
        return self.window_seq

    @property
    def centers(self):
        return self.window_seq.keys()

    @inplacify
    def add_window_generic(self, array, **kwargs):
        """
        Build a window sequence of given parameters with centers
        specified with ``array`` argument.

        Parameters
        ----------
        array : list, np.ndarray
            The array containing the centers of windows.
        kwargs : dict
            Keyword arguments to pass to the `window_class`.
        """
        if not isinstance(array, (list, np.ndarray)):
            raise TypeError("Expected list-like as ``array``.")
        for center in array:
            self.add_window(center=center, **kwargs)
        return self

    @inplacify
    def add_window_arange(self, start, stop, step, **kwargs):
        """
        Build a window sequence of given parameters to apply on ifg.
        Works similar to numpy.arange.

        Parameters
        ----------
        start : float
            The start of the centers.
        stop : float
            The end value of the center
        step : float
            The step value to increment center.
        kwargs : dict
            Keyword arguments to pass to the `window_class`.
        """
        arr = np.arange(start, stop, step)
        for cent in arr:
            self.add_window(center=cent, **kwargs)
        return self

    @inplacify
    def add_window_linspace(self, start, stop, num, **kwargs):
        """
        Build a window sequence of given parameters to apply on ifg.
        Works similar to numpy.linspace.

        Parameters
        ----------
        start : float
            The start of the centers.
        stop : float
            The end value of the center
        num : float
            The number of Gaussian windows.
        kwargs : dict
            Keyword arguments to pass to the `window_class`.
        """
        arr = np.linspace(start, stop, num)
        for cent in arr:
            self.add_window(center=cent, **kwargs)
        return self

    @inplacify
    def add_window_geomspace(self, start, stop, num, **kwargs):
        """
        Build a window sequence of given parameters to apply on ifg.
        Works similar to numpy.geomspace.

        Parameters
        ----------
        start : float
            The start of the centers.
        stop : float
            The end value of the center
        num : float
            The number of Gaussian windows.
        kwargs : dict
            Keyword arguments to pass to the `window_class`.
        """
        arr = np.geomspace(start, stop, num)
        for cent in arr:
            self.add_window(center=cent, **kwargs)
        return self

    def view_windows(self, ax=None, maxsize=80, **kwargs):
        """
        Gives a rough view of the different windows along with the ifg.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            An axis to draw the plot on. If not given, it will plot
            of the last used axis.
        maxsize : int, optional
            The maximum number of Gaussian windows to display on plot.
            Default is 80, but be aware that setting a high value can
            drastically reduce performance.
        kwargs : dict, optional
            Additional keyword arguments to pass to plot function.
        """
        winlen = len(self.window_seq)
        if maxsize != 0:
            ratio = winlen // maxsize
            if winlen > maxsize:
                warnings.warn(
                    "Image seems crowded, displaying only a subsample of the given windows.",
                    PySprintWarning
                )
                for i, (_, val) in enumerate(self.window_seq.items()):
                    if i % ratio == 0:
                        val.plot(ax=ax, scalefactor=np.max(self.y) * .75, **kwargs)
            else:
                for _, val in self.window_seq.items():
                    val.plot(ax=ax, scalefactor=np.max(self.y) * .75, **kwargs)
        self.plot(ax=ax)

    @inplacify
    def remove_all_windows(self):
        """
        Remove all the Gaussian windows.
        """
        self.window_seq.clear()
        return self

    @inplacify
    def reset_state(self):
        """
        Reset the object's state fully: delete all the
        calculated GD, caches, heatmaps and window sequences.
        """
        self.remove_all_windows()
        self.found_centers.clear()
        self.X_cont = np.array([])
        self.Y_cont = np.array([])
        self.Z_cont = np.array([])
        self.GD = None
        self.cachedlen = 0
        self.fastmath = True
        return self

    @inplacify
    def remove_window_at(self, center):
        """
        Removes a window at center.

        Parameters
        ----------
        center : float
            The center of the window to remove.
            Raises ValueError if there is not such window.
        """
        if center not in self.window_seq.keys():
            c = find_nearest(
                np.fromiter(self.window_seq.keys(), dtype=float), center
            )
            raise ValueError(
                f"There is no window with center {center}. "
                f"Did you mean {c[0]}?"
            )
        self.window_seq.pop(center, None)
        return self

    @inplacify
    def remove_window_interval(self, start, stop):
        """
        Remove window interval inclusively.

        Parameters
        ----------
        start : float
            The start value of the interval.
        stop : float
            The stop value of the interval.
        """
        wins = np.fromiter(self.window_seq.keys(), dtype=float)
        mask = wins[(wins <= stop) & (wins >= start)]
        for center in mask:
            self.window_seq.pop(center, None)
        return self

    @inplacify
    def cover(self, N, **kwargs):
        """
        Cover the whole domain with `N` number of windows
        uniformly built with the given parameters.

        Parameters
        ----------
        N : float
            The number of Gaussian windows.
        kwargs : dict
            Keyword arguments to pass to the `window_class`.
        """

        self.add_window_linspace(np.min(self.x), np.max(self.x), N, **kwargs)

    def calculate(
            self,
            reference_point,
            order,
            show_graph=False,
            silent=False,
            force_recalculate=False,
            fastmath=True,
            usenifft=False,
            parallel=False,
            ransac=False,
            errors="ignore",
            **kwds
    ):
        """
        Calculates the dispersion.

        Parameters
        ----------
        reference_point : float
            The reference point.
        order : int
            The dispersion order to look for. Must be in [2, 6].
        show_graph : bool, optional
            Whether to show the GD graph on complete. Default is False.
        silent : bool, optional
            Whether to print progressbar. By default it will print.
        force_recalculate : bool, optional
            Force to recalculate the GD graph not only the curve fitting.
            Default is False.
        fastmath : bool, optional
            Whether to build additional arrays to display heatmap.
            Default is True.
        usenifft : bool, optional
            Whether to use Non-unfirom FFT when calculating GD.
            Default is False. **Not stable.**
        parallel : bool, optional
            Whether to use parallel computation. Only availabe if `Dask`
            is installed. The speedup is about 50-70%. Default is False.
        ransac : bool, optional
            Whether to use RANSAC filtering on the detected peaks. Default
            is False.
        errors : str, optional
            Whether to raise an error is the algorithm couldn't find the
            center of the peak. Default is "ignore".
        kwds : optional
            Other keyword arguments to pass to RANSAC filter.

        Raises
        ------
        ValueError, if no window sequence is added to the interferogram.
        ValueError, if order is 1.
        ModuleNotFoundError, if `Dask` is not available when using parallel=True.
        """
        if len(self.window_seq) == 0:
            raise ValueError("Before calculating a window sequence must be set.")

        if self.cachedlen != len(self.window_seq) or fastmath != self.fastmath:
            force_recalculate = True
        self.fastmath = fastmath
        if force_recalculate:
            self.found_centers.clear()
            self.build_GD(
                silent=silent, fastmath=fastmath, usenifft=usenifft, parallel=parallel, errors=errors
            )
        if self.GD is None:
            self.build_GD(
                silent=silent, fastmath=fastmath, usenifft=usenifft, parallel=parallel, errors=errors
            )

        self.cachedlen = len(self.window_seq)

        if order == 1 or order > 6:
            raise ValueError("Order must be in [2, 6].")

        if ransac:
            print("Running RANSAC-filter..")
            self.GD.ransac_filter(order=order, plot=show_graph, **kwds)
            self.GD.apply_filter()

        d, ds, fr = self.GD._fit(
            reference_point=reference_point, order=order
        )
        if show_graph:
            self.GD.plot()
        return d, ds, fr

    def build_GD(self, silent=False, fastmath=True, usenifft=False, parallel=False, errors="ignore"):
        """
        Build the GD.

        Parameters
        ----------
        silent : bool, optional
            Whether to print progressbar. By default it will print.
        fastmath : bool, optional
            Whether to build additional arrays to display heatmap.
            Default is True.
        usenifft : bool, optional
            Whether to use Non-unfirom FFT when calculating GD.
            Default is False. **Not stable.**
        parallel : bool, optional
            Whether to use parallel computation. Only availabe if `Dask`
            is installed. The speedup is about 50-70%. Default is False.
        errors : str, optional
            Whether to raise an error is the algorithm couldn't find the
            center of the peak.

        Returns
        -------
        GD : pysprint.core.phase.Phase
            The phase object with `GD_mode=True`. See its docstring for more info.
        """
        if parallel:

            if not CAN_PARALLELIZE:
                raise ModuleNotFoundError(
                    "Module `dask` not found. Please install it in order to use parallelism."
                )

            else:
                self.fastmath = fastmath
                self._apply_window_seq_parallel(fastmath=fastmath, usenifft=usenifft, errors=errors)

                if not silent:
                    with ProgressBar():
                        computed = compute(*self.found_centers.values())

                else:
                    computed = compute(*self.found_centers.values())

                cleaned_delays = [
                    k for i, k in enumerate(self.found_centers.keys()) if computed[i] is not None
                ]
                delay = np.fromiter(cleaned_delays, dtype=float)

                omega = np.fromiter([c for c in computed if c is not None], dtype=float)

                if not silent:
                    print(f"Skipped: {len(self.window_seq) - sum(1 for _ in filter(None.__ne__, computed))}")

        else:
            self.fastmath = fastmath
            self._apply_window_sequence(silent=silent, fastmath=fastmath, usenifft=usenifft)
            self._clean_centers()

            delay = np.fromiter(self.found_centers.keys(), dtype=float)
            omega = np.fromiter(self.found_centers.values(), dtype=float)

        self.GD = Phase(delay, omega, GD_mode=True)
        return self.GD

    def build_phase(self):
        raise NotImplementedError("Use `build_GD` instead.")

    def _predict_ideal_window_fwhm(self):
        pass

    def _apply_window_sequence(
            self, silent=False, fastmath=True, usenifft=False, errors="ignore"
    ):
        winlen = len(self.window_seq)
        self.errorcounter = 0
        if not fastmath:
            # here we setup the shape for the Z array because
            # it is much faster than using np.append in every iteration
            _x, _y, _, _ = self._safe_cast()
            _obj = FFTMethod(_x, _y)
            _obj.ifft(usenifft=usenifft)
            x, y = find_roi(_obj.x, _obj.y)
            self.Y_cont = np.array(x)
            yshape = y.size
            xshape = len(self.window_seq)
            self.Z_cont = np.empty(shape=(yshape, xshape))

        for idx, (_center, _window) in enumerate(self.window_seq.items()):
            _x, _y, _, _ = self._safe_cast()
            _obj = FFTMethod(_x, _y)
            _obj.y *= _window.y
            _obj.ifft(usenifft=usenifft)
            x, y = find_roi(_obj.x, _obj.y)
            if not fastmath:
                self.Z_cont[:, idx] = y
            try:
                centx, _ = find_center(x, y)
                self.found_centers[_center] = centx
            except ValueError as err:
                self.errorcounter += 1
                if errors == "ignore":
                    self.found_centers[_center] = None
                else:
                    raise err
            if not silent:  # This creates about 5-15% overhead.. maybe create a buffer
                sys.stdout.write('\r')
                j = (idx + 1) / winlen
                sys.stdout.write(
                    "Progress : [%-30s] %d%% (Skipped: %d)" % ('=' * int(30 * j), 100 * j, self.errorcounter)
                )
                sys.stdout.flush()

    def _apply_window_seq_parallel(
            self, fastmath=True, usenifft=False, errors="ignore"
    ):
        self.errorcounter = 0
        if not fastmath:
            # here we setup the shape for the Z array and allocate Y, because
            # it is much faster than using np.append in every iteration
            _x, _y, _, _ = self._safe_cast()
            _obj = FFTMethod(_x, _y)
            _obj.ifft(usenifft=usenifft)
            x, y = find_roi(_obj.x, _obj.y)
            yshape = y.size
            self.Y_cont = np.array(x)
            xshape = len(self.window_seq)
            self.Z_cont = np.empty(shape=(yshape, xshape))

        for idx, (_center, _window) in enumerate(self.window_seq.items()):
            element = self._prepare_element(idx, _window, fastmath, usenifft, errors)
            if element is None:
                self.errorcounter += 1  # This might be useless, since we lazy evaluate things..
            self.found_centers[_center] = element

    @delayed
    def _prepare_element(self, idx, window, fastmath=True, usenifft=False, errors="ignore"):
        _x, _y, _, _ = self._safe_cast()
        _obj = FFTMethod(_x, _y)
        _obj.y *= window.y
        _obj.ifft(usenifft=usenifft)
        x, y = find_roi(_obj.x, _obj.y)
        if not fastmath:
            self.Z_cont[:, idx] = y
        try:
            centx, _ = find_center(x, y)
            return centx
        except ValueError as err:
            if errors == "ignore":
                return None
            else:
                raise err

    def _clean_centers(self, silent=False):
        dct = {k: v for k, v in self.found_centers.items() if v is not None}
        self.found_centers = dct
        winlen = len(self.window_seq)
        usefullen = len(self.found_centers)

        if not silent:
            if winlen != usefullen:
                print(
                    f"\n{abs(winlen-usefullen)} points skipped "
                    f"due to ambiguous peak position."
                )

    def errorplot(self, *args, **kwargs):
        """
        Plot the errors of fitting.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            An axis to draw the plot on. If not given, it will plot
            of the last used axis.
        percent : bool, optional
            Whether to plot percentage difference. Default is False.
        title : str, optional
            The title of the plot. Default is "Errors".
        kwargs : dict, optional
            Additional keyword arguments to pass to plot function.
        """
        try:
            getattr(self.GD, "errorplot", None)(*args, **kwargs)
        except TypeError:
            raise NotCalculatedException("Must calculate before plotting errors.")

    @property
    def get_GD(self):
        """
        Return the GD if it is already calculated.
        """
        if self.GD is not None:
            return self.GD
        raise NotCalculatedException("Must calculate GD first.")

    @property
    def errors(self):
        """
        Return the fitting errors as np.ndarray.
        """
        return getattr(self.GD, "errors", None)

    def _collect_failures(self):
        return [k for k in self.window_seq.keys() if k not in self.found_centers.keys()]

    def _construct_heatmap_data(self):
        self.X_cont = np.fromiter(self.window_seq.keys(), dtype=float)

    def heatmap(self, ax=None, levels=None, cmap="viridis", include_ridge=True):
        """
        Plot the heatmap.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            An axis to draw the plot on. If not given, it will plot
            of the last used axis.
        levels : np.ndarray, optional
            The levels to use for plotting.
        cmap : str, optional
            The colormap to use.
        include_ridge : bool, optional
            Whether to mark the detected ridge of the plot.
            Default is True.
        """
        if self.GD is None:
            raise NotCalculatedException("Must calculate GD first.")

        if self.fastmath:
            raise ValueError(
                "You need to recalculate with `fastmath=False` to plot the heatmap."
            )
        # Only construct if we need to..
        if not (self.Y_cont.size, self.X_cont.size) == self.Z_cont.shape:
            self._construct_heatmap_data()
        if ax is None:
            plt.contourf(
                self.X_cont, self.Y_cont, self.Z_cont, levels=levels, cmap=cmap, extend="both"
            )
        else:
            ax.contourf(
                self.X_cont, self.Y_cont, self.Z_cont, levels=levels, cmap=cmap, extend="both"
            )
        if include_ridge:
            if ax is None:
                plt.plot(*self.GD.data, color='red', label='detected ridge')
            else:
                ax.plot(*self.GD.data, color='red', label='detected ridge')
            plt.legend()
        if ax is None:
            plt.xlabel('Window center [PHz]')
            plt.ylabel('Delay [fs]')
            try:
                upper_bound = min(1.5 * np.max(self.GD.data[1]), np.max(self.Y_cont))
                plt.ylim(None, upper_bound)
            except ValueError:
                pass
            # ValueError: zero-size array to reduction operation maximum which has no identity
            # This means that the array is empty, we should pass that case.
        else:
            ax.set_autoscalex_on(False)
            try:
                upper_bound = min(1.5 * np.max(self.GD.data[1]), np.max(self.Y_cont))
                ax.set(
                    xlabel="Window center [PHz]",
                    ylabel="Delay [fs]",
                    ylim=(None, upper_bound)
                )
            except ValueError:
                pass

    def get_heatmap_data(self):
        """
        Return the data which was used to create the heatmap.

        Returns
        -------
        X_cont : np.ndarray
            The window centers with shape (n,).
        Y_cont : np.ndarray
            The time axis calculated from the IFFT of the dataset with shape (m,).
        Z_cont : np.ndarray
            2D array with shape (m, n) containing the depth information.
        """
        if all([self.Y_cont.size != 0, self.Z_cont.size != 0]):
            self._construct_heatmap_data()
        else:
            raise ValueError("Must calculate with `fastmath=False` before trying to access the heatmap data.")

        return self.X_cont, self.Y_cont, self.Z_cont
